import os
import sys


REPO_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
EXAMPLES_DIR = os.path.join(REPO_DIR, "examples")
EXAMPLESCRAPER_DIR = os.path.join(EXAMPLES_DIR, "examplescraper")

# Enable `examplescraper` lookup as tests leverage it.
sys.path.insert(0, EXAMPLES_DIR)



import os
import sys
from time import sleep

from requests import get, RequestException


if __name__ == "__main__":
    host = os.environ.get("SERVICE_NAME_MASTER_SCRAPER")
    port = sys.argv[1]
    url = "http://{host}:{port}/health/".format(host=host, port=port)

    i = 0
    while i < 10:
        response = None

        try:
            response = get(url)
        except RequestException:
            pass

        if response and response.status_code == 200:
            break
        else:
            i += 1
            sleep(0.5)

    if i > 10:
        sys.exit("Healthcheck failed.")



class Config:
    #
    # General settings.
    # Where to place the *.sqlite DB file populated with scraped data.
    DATA_DIRECTORY = "/tmp"

    # Log level settings.
    # NOTE 'DEBUG' will be very verbose.
    LOG_LEVEL = "WARNING"

    # How often data (how many items at once) should be commited to the DB.
    TRANSACTION_SIZE = 5000

    #
    # Scraping settings.
    # Number of threads used to asynchronously scrape data from URLs.
    WORKERS_COUNT = 50

    # How long to wait for a response (in seconds).
    REQUEST_TIMEOUT = 10

    # User agents to use in requests.
    # NOTE must be populated before starting the scraping process.
    USER_AGENTS = None

    # =========================================================================
    # DEFAULT SERVICES SETTINGS: MUST MATCH HOST SETTINGS.
    # =========================================================================

    #
    # Tor.
    TOR_PORT = 9051
    TOR_PASSWORD = "I-solemnly-swear-I-am-up-to-no-good"

    #
    # Privoxy.
    PRIVOXY_PORT = 8118
    PRIVOXY_HOST = "127.0.0.1"

    #
    # TorIpChanger.
    LOCAL_HTTP_PROXY = "{0}:{1}".format(PRIVOXY_HOST, PRIVOXY_PORT)
    NEW_IP_MAX_ATTEMPTS = 1000
    REUSE_THRESHOLD = 10
    TORIPCHANGER_CLASS = (
        "scrapemeagain.dockerized.ipchanger.DockerizedTorIpChanger"
    )

    # =========================================================================
    # DOCKERIZED SETTINGS.
    # =========================================================================

    DOCKER_INTERFACE_NAME = "docker0"

    #
    # Orchestration.
    SCRAPERS_COUNT = 1

    #
    # IpStore.
    IPSTORE_PORT = 5000
    IPSTORE_REUSE_THRESHOLD = REUSE_THRESHOLD * SCRAPERS_COUNT

    #
    # URLBroker.
    URLBROKER_PORT = 6000
    # NOTE each scraper MUST set a custom `ListUrlsBroker` subclass in
    # 'scrapemeagain.scrapers.{your scraper}.config.URLBROKER_CLASS'
    URLBROKER_CLASS = (
        "scrapemeagain.dockerized.apps.urlbroker.urlbrokers.ListUrlsBroker"
    )

    #
    # DataStore.
    DATASTORE_PORT = 7000
    # NOTE set 'scrapemeagain.scrapers.{your scraper}.config.DATASTORE_CLASS'
    # if your scraper adds custom functionality to `DataStoreDatabaser`.
    DATASTORE_DATABASER_CLASS = "scrapemeagain.databaser.DataStoreDatabaser"

    #
    # Healthcheck.
    HEALTHCHECK_PORT = 8000



from flask import jsonify
from toripchanger import TorIpChanger

from scrapemeagain.config import Config
from scrapemeagain.dockerized.utils import app_factory


app = app_factory(__name__)


# Global IP store (using only specific `TorIpChanger` functionality).
IPSTORE = TorIpChanger(reuse_threshold=Config.IPSTORE_REUSE_THRESHOLD)


@app.route("/ip-is-safe/<ip>/")
def ip_is_safe(ip):
    safe = IPSTORE._ip_is_safe(ip)
    if safe:
        IPSTORE._manage_used_ips(ip)

    return jsonify({"safe": safe})


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=Config.IPSTORE_PORT)



from flask import jsonify

from scrapemeagain.config import Config
from scrapemeagain.dockerized.utils import app_factory, get_class_from_path


app = app_factory(__name__)


urlbroker_class = get_class_from_path(Config.URLBROKER_CLASS)
URLBROKER = urlbroker_class()


@app.route("/list-urls-range/")
def list_urls_range():
    start, end = URLBROKER.get_urls_range()
    return jsonify({"start": start, "end": end})


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=Config.URLBROKER_PORT)



import unittest
from unittest.mock import patch

from tests.utils import import_docker_compose


docker_compose = import_docker_compose()


EXPECTED_EXAMPLESCRAPER_COMPOSE = """
version: '3'
services:
  examplescraper-scp1:
    image: dusanmadar/scrapemeagain:1.0.5
    entrypoint: /scrapemeagain/scrapemeagain/dockerized/entrypoints/entrypoint.scp1.sh
    volumes:
      - /tmp/examplescraper:/scp/examplescraper
    environment:
      - DOCKER_HOST_IP=fake_docker_host_ip
      - SCRAPER_PACKAGE=examplescraper
      - SCRAPER_CONFIG=tests.integration.fake_config
      - SERVICE_NAME_TEMPLATE=examplescraper-scp{}
      - SERVICE_NAME_MASTER_SCRAPER=examplescraper-scp1
      - TOR_PORT=9051
      - TOR_PASSWORD=I-solemnly-swear-I-am-up-to-no-good
      - PRIVOXY_PORT=8118
      - PRIVOXY_HOST=127.0.0.1
      - IPSTORE_PORT=5000
      - URLBROKER_PORT=6000
      - DATASTORE_PORT=7000
      - HEALTHCHECK_PORT=8000
  examplescraper-scp2:
    image: dusanmadar/scrapemeagain:1.0.5
    entrypoint: /scrapemeagain/scrapemeagain/dockerized/entrypoints/entrypoint.scpx.sh
    volumes:
      - /tmp/examplescraper:/scp/examplescraper
    environment:
      - DOCKER_HOST_IP=fake_docker_host_ip
      - SCRAPER_PACKAGE=examplescraper
      - SCRAPER_CONFIG=tests.integration.fake_config
      - SERVICE_NAME_TEMPLATE=examplescraper-scp{}
      - SERVICE_NAME_MASTER_SCRAPER=examplescraper-scp1
      - TOR_PORT=9051
      - TOR_PASSWORD=I-solemnly-swear-I-am-up-to-no-good
      - PRIVOXY_PORT=8118
      - PRIVOXY_HOST=127.0.0.1
      - IPSTORE_PORT=5000
      - URLBROKER_PORT=6000
      - DATASTORE_PORT=7000
      - HEALTHCHECK_PORT=8000
""".strip()


class DockerComposeTestCase(unittest.TestCase):
    @patch("scrapemeagain-compose.DOCKER_HOST_IP", "fake_docker_host_ip")
    def test_construct_compose_file(self):
        """
        Test `construct_compose_file` returns expected yaml.
        """
        self.assertEqual(
            EXPECTED_EXAMPLESCRAPER_COMPOSE,
            docker_compose.construct_compose_file(
                "/tmp/examplescraper", "tests.integration.fake_config"
            ),
        )

    def test_construct_compose_file_no_extra_config(self):
        """
        Test `construct_compose_file` returns expected yaml without
        `SCRAPER_CONFIG` env when no extra config is provided.
        """
        yml = docker_compose.construct_compose_file("/tmp/examplescraper")
        self.assertNotIn("SCRAPER_CONFIG", yml)

    def test_construct_compose_file_nonexisting_scraper(self):
        """
        Test `construct_compose_file` raises `ModuleNotFoundError` for a
        nonexisting scraper.
        """
        with self.assertRaises(ModuleNotFoundError):
            docker_compose.construct_compose_file("/nonexisting")



"""Global logging setup."""


import logging
import os

import requests

from scrapemeagain.config import Config
from scrapemeagain.utils.alnum import get_current_date


def setup_logging(logger_name):
    # Ensure log dir exists.
    log_dir = os.path.join(Config.DATA_DIRECTORY, "log")
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    # Create/clean log file.
    file_name = "{0}_{1}.log".format(logger_name, get_current_date())
    log_file = os.path.join(log_dir, file_name)
    if os.path.exists(log_file):
        os.remove(log_file)

    # Set log level.
    log_level = getattr(logging, Config.LOG_LEVEL, None)
    if not isinstance(log_level, int):
        raise ValueError('Invalid log level: "{}"'.format(Config.LOG_LEVEL))

    # Configure logger.
    logging.basicConfig(
        filename=log_file,
        format="%(levelname)s %(asctime)s - %(message)s",
        level=log_level,
    )

    # Silence `requests` and `stem` modules logging.
    requests_log = logging.getLogger("requests")
    requests_log.setLevel(logging.WARNING)
    requests.packages.urllib3.disable_warnings()

    stem_log = logging.getLogger("stem")
    stem_log.setLevel(logging.WARNING)



from examplescraper.config import Config

Config.WORKERS_COUNT = 5

# `examplesite` runs locally so we cannot route traffic via privoxy and tor.
Config.LOCAL_HTTP_PROXY = ""

# Don't actualy change tor's IP address.
Config.TORIPCHANGER_CLASS = (
    "tests.integration.fake_ipchanger.FakeDockerizedTorIpChanger"
)

Config.DATA_DIRECTORY = "/tmp/test-examplescraper"



#!/usr/bin/env python3


"""
Dynamic `docker-compose.yml` based on settings defined in `config`.

Usage:
    `python3 scrapemeagain-compose.py -s <scraper> [-c <path.to.some.config>] | docker-compose -f - up`

    Example:
    $ cd ../
    $ python3 scripts/scrapemeagain-compose.py -s $(pwd)/examples/examplescraper | docker-compose -f - up
    $ python3 scripts/scrapemeagain-compose.py -s $(pwd)/examples/examplescraper -c tests.integration.fake_config | docker-compose -f - up
"""  # noqa


import argparse
import inspect
import os
import sys

from jinja2 import Template

from scrapemeagain.config import Config
from scrapemeagain.dockerized.utils import (
    apply_scraper_config,
    get_inf_ip_address,
)

# Must match Dockerfile's `SCP_DIR` env variable.
SCP_DIR = "/scp"
# Must match Dockerfile's `APP_SRC_DIR` env variable.
APP_SRC_DIR = "/scrapemeagain/scrapemeagain"
ENTRYPOINT_DIR = os.path.join(APP_SRC_DIR, "dockerized", "entrypoints")
ENTRYPOINT_TEMPLATE = os.path.join(ENTRYPOINT_DIR, "entrypoint.scp{}.sh")
DOCKER_HOST_IP = get_inf_ip_address(Config.DOCKER_INTERFACE_NAME)


def create_scraper_service(id, package, path, config):
    service_name_template = "{}-scp{{}}".format(package)

    return {
        "name": service_name_template.format(id),
        "entrypoint": ENTRYPOINT_TEMPLATE.format(1 if id == 1 else "x"),
        "volumes": ["{}:{}".format(path, os.path.join(SCP_DIR, package))],
        "environment": {
            "DOCKER_HOST_IP": DOCKER_HOST_IP,
            "SCRAPER_PACKAGE": package,
            "SCRAPER_CONFIG": config,
            "SERVICE_NAME_TEMPLATE": service_name_template,
            # The first scraper service is always the master one.
            "SERVICE_NAME_MASTER_SCRAPER": service_name_template.format(1),
            "TOR_PORT": Config.TOR_PORT,
            "TOR_PASSWORD": Config.TOR_PASSWORD,
            "PRIVOXY_PORT": Config.PRIVOXY_PORT,
            "PRIVOXY_HOST": Config.PRIVOXY_HOST,
            "IPSTORE_PORT": Config.IPSTORE_PORT,
            "URLBROKER_PORT": Config.URLBROKER_PORT,
            "DATASTORE_PORT": Config.DATASTORE_PORT,
            "HEALTHCHECK_PORT": Config.HEALTHCHECK_PORT,
        },
    }


def construct_compose_file(path, config=None):
    # Apply scraper specific config (but first ensure it's accessible).
    sys.path.insert(0, os.path.dirname(path))
    package = path.split(os.sep)[-1]
    apply_scraper_config(package, config)

    services = [
        create_scraper_service(id, package, path, config)
        for id in range(1, Config.SCRAPERS_COUNT + 1)
    ]

    # A dynamic way to get the path of `scrapemeagain.dockerized` package.
    # Required because this file must be runnable both during development and
    # after installing the package.
    # Credits: https://stackoverflow.com/q/2419416/4183498.
    dockerized_utils_module_path = inspect.getsourcefile(apply_scraper_config)
    template_dir = os.path.dirname(dockerized_utils_module_path)
    with open(os.path.join(template_dir, "docker-compose.yml")) as f:
        template = Template(f.read())

    return template.render(services=services).strip()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-s",
        "--scraper",
        required=True,
        help=(
            "Absolute path to scraper package dir; e.g. '/tmp/examplescraper'."
        ),
    )
    parser.add_argument(
        "-c",
        "--config",
        required=False,
        help=(
            "Dotted path to a specific (not 'config.py') config module; "
            "e.g. 'examplescraper.more_configs.test_config'. The path must be"
            "importbale locally and in the container."
        ),
        default=None,
    )

    args = parser.parse_args()
    print(construct_compose_file(args.scraper, args.config))


if __name__ == "__main__":
    main()



"""SQLAlchemy common database tables definition."""


from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base


Base = declarative_base()


class ItemUrlsTable(Base):
    """Item URLs table structure."""

    __tablename__ = "item_urls"

    id = Column(Integer, primary_key=True)
    url = Column(String)

    def __repr__(self):
        """Nice ItemUrlsTable row representation."""
        return "<Item URL (id={id}, url={url})>".format(
            id=self.id, url=self.url
        )



from unittest.mock import patch, PropertyMock

from requests import Response

from tests.pipeline_base import TestPipelineBase
from scrapemeagain.pipeline import EXIT, DUMP_URLS_BUCKET
from scrapemeagain.utils.http import get


def create_responses(mock_urls, mock_statuses):
    responses = []

    for mock_url, mock_status in zip(mock_urls, mock_statuses):
        response = Response()
        response.url = mock_url
        response.status_code = mock_status

        responses.append(response)

    return responses


class TestPipeline(TestPipelineBase):
    @patch("scrapemeagain.pipeline.Value")
    @patch("scrapemeagain.pipeline.Event")
    @patch("scrapemeagain.pipeline.ThreadPoolExecutor")
    @patch("scrapemeagain.pipeline.Queue")
    def test_prepare_pipeline(
        self, mock_queue, mock_pool, mock_event, mock_value
    ):
        """Test 'prepare_pipeline' initializes all necessary
        multithreading and multiprocessing objects.
        """
        self.pipeline.prepare_pipeline()

        self.assertTrue(mock_queue.call_count, 3)
        mock_pool.assert_called_once_with(self.pipeline.workers_count)
        self.assertTrue(mock_event.call_count, 2)
        self.assertTrue(mock_value.call_count, 2)

    @patch("scrapemeagain.pipeline.get_current_datetime")
    @patch("scrapemeagain.pipeline.logging")
    @patch("scrapemeagain.pipeline.print", create=True)
    def test_inform(self, mock_print, mock_logging, mock_get_current_datetime):
        """Test 'inform' is able to both log and print a message."""
        mock_get_current_datetime.return_value = "a datetime"
        mock_message = "something happened"

        self.pipeline.inform(mock_message)

        mock_logging.info.assert_called_once_with(mock_message)
        mock_print.assert_called_once_with(
            "{0} {1}".format("a datetime", "something happened"), end="\n"
        )

    @patch("scrapemeagain.pipeline.Pipeline.inform")
    def test_inform_progress(self, mock_inform):
        """Test '_inform_progress' prints scraping progress."""
        self.pipeline.urls_processed.value = 10
        self.pipeline.urls_to_process.value = 100

        self.pipeline._inform_progress()

        mock_inform.assert_called_once_with(
            "Processed 10 (10.00%) URLs", log=False, end="\r"
        )

    @patch("scrapemeagain.pipeline.Pipeline.inform")
    def test_inform_progress_fail(self, mock_inform):
        """Test '_inform_progress' does nothing on ZeroDivisionError."""
        self.pipeline.urls_to_process.value = 0

        self.pipeline._inform_progress()

        with self.assertRaises(AssertionError):
            mock_inform.assert_called_once_with()

    @patch("scrapemeagain.pipeline.logging")
    def test_change_ip(self, mock_logging):
        """Test 'change_ip' simply sets a new IP via Tor."""
        self.pipeline.tor_ip_changer.get_new_ip.return_value = "8.8.8.8"

        self.pipeline.change_ip()

        self.pipeline.tor_ip_changer.get_new_ip.assert_called_once_with()
        mock_logging.info.assert_called_once_with(
            "New IP: {new_ip}".format(new_ip="8.8.8.8")
        )

    @patch("scrapemeagain.pipeline.logging")
    def test_change_ip_fail(self, mock_logging):
        """Test 'change_ip' tries again on fail."""
        self.pipeline.tor_ip_changer.get_new_ip.side_effect = [
            ValueError,
            "8.8.8.8",
        ]

        self.pipeline.change_ip()

        self.assertTrue(self.pipeline.tor_ip_changer.get_new_ip.call_count, 2)
        mock_logging.error.assert_called_once_with("Failed setting new IP")
        mock_logging.info.assert_called_once_with(
            "New IP: {new_ip}".format(new_ip="8.8.8.8")
        )

    @patch("scrapemeagain.pipeline.time.sleep")
    @patch("scrapemeagain.pipeline.Pipeline.inform")
    def test_produce_urls_wrapper(self, mock_inform, mock_sleep):
        """Test '_produce_urls_wrapper' informs about URLs count and manages
        the 'producing_urls_in_progress' event.
        """
        self.pipeline._produce_urls_wrapper(lambda: 5)

        self.assertEqual(self.pipeline.urls_to_process.value, 5)
        mock_inform.assert_called_once_with(
            "URLs to process: {}".format(self.pipeline.urls_to_process.value)
        )
        mock_sleep.assert_called_once_with(1)
        self.pipeline.producing_urls_in_progress.set.assert_called_once_with()
        self.pipeline.producing_urls_in_progress.clear.assert_called_once_with()

    @patch("scrapemeagain.pipeline.Pipeline.inform")
    @patch("scrapemeagain.pipeline.time.sleep")
    def test_produce_list_urls(self, mock_sleep, mock_inform):
        """Test 'produce_list_urls' uses 'scraper.generate_list_urls'."""
        mock_urls = ["url1", "url2", "url3"]
        self.pipeline.scraper.generate_list_urls.return_value = mock_urls

        self.pipeline.produce_list_urls()

        self.pipeline.scraper.generate_list_urls.assert_called_once_with()
        for mock_url in mock_urls:
            self.pipeline.url_queue.put.assert_any_call(mock_url)

    @patch("scrapemeagain.pipeline.Pipeline.inform")
    @patch("scrapemeagain.pipeline.time.sleep")
    def test_produce_item_urls(self, mock_sleep, mock_inform):
        """Test 'produce_item_urls' uses 'databaser.get_item_urls'."""
        mock_urls = [("url1",), ("url2",), ("url3",)]
        self.pipeline.databaser.get_item_urls.return_value = mock_urls

        self.pipeline.produce_item_urls()

        self.pipeline.databaser.get_item_urls.assert_called_once_with()
        for mock_url in mock_urls:
            self.pipeline.url_queue.put.assert_any_call(mock_url[0])

    def test_classify_response_ok(self):
        """Test '_classify_response' puts an OK response to
        'response_queue'."""
        mock_response_ok = Response()
        mock_response_ok.url = "url1"
        mock_response_ok.status_code = 200

        self.pipeline._classify_response(mock_response_ok)

        self.pipeline.response_queue.put.assert_called_once_with(
            mock_response_ok
        )
        with self.assertRaises(AssertionError):
            self.pipeline.url_queue.put.assert_called_once_with(
                mock_response_ok.url
            )

    def test_classify_response_not_ok(self):
        """Test '_classify_response' puts a non OK response URL back to
        'url_queue'."""
        mock_response_not_ok = Response()
        mock_response_not_ok.url = "url1"
        mock_response_not_ok.status_code = 500

        self.pipeline._classify_response(mock_response_not_ok)

        self.pipeline.url_queue.put.assert_called_once_with(
            mock_response_not_ok.url
        )
        with self.assertRaises(AssertionError):
            self.pipeline.response_queue.put.assert_called_once_with(
                mock_response_not_ok
            )

    @patch("scrapemeagain.pipeline.logging")
    @patch("scrapemeagain.pipeline.Pipeline._classify_response")
    def test_actually_get_html(self, mock_classify_response, mock_logging):
        """Test '_actually_get_html' fires requests for given URLs."""
        mock_urls = ["url1", "url2", "url3"]
        mock_url_statuses = [200, 500, 200]
        self.pipeline.pool.map.return_value = create_responses(
            mock_urls, mock_url_statuses
        )

        self.pipeline._actually_get_html(mock_urls)

        self.pipeline.pool.map.assert_called_once_with(get, mock_urls)
        self.assertEqual(mock_classify_response.call_count, len(mock_urls))
        self.pipeline.requesting_in_progress.set.assert_called_once_with()
        self.pipeline.requesting_in_progress.clear.assert_called_once_with()
        with self.assertRaises(AssertionError):
            mock_logging.error.assert_called_once_with("Failed scraping URLs")

    @patch("scrapemeagain.pipeline.logging")
    def test_actually_get_html_fails(self, mock_logging):
        """Test '_actually_get_html' logs an exception message on fail."""
        self.pipeline.pool.map.side_effect = ValueError

        self.pipeline._actually_get_html(None)

        self.pipeline.pool.map.assert_called_once_with(get, None)
        self.pipeline.requesting_in_progress.set.assert_called_once_with()
        self.pipeline.requesting_in_progress.clear.assert_called_once_with()
        mock_logging.error.assert_called_once_with("Failed scraping URLs")
        self.assertEqual(mock_logging.exception.call_count, 1)

    @patch("scrapemeagain.pipeline.Pipeline._actually_get_html")
    @patch("scrapemeagain.pipeline.Pipeline.change_ip")
    def test_get_html(self, mock_change_ip, mock_actually_get_html):
        """Test 'get_html' populates and passes URL bulks for processing."""
        mock_urls = ["url1", "url2", "url3"]
        self.pipeline.url_queue.get.side_effect = (
            mock_urls + [DUMP_URLS_BUCKET] + [EXIT]
        )
        self.pipeline.workers_count = 4

        self.pipeline.get_html()

        # 4 = len(mock_urls) + DUMP_URLS_BUCKET + EXIT
        self.assertEqual(self.pipeline.url_queue.get.call_count, 5)

        mock_change_ip.assert_called_once_with()
        mock_actually_get_html.assert_called_once_with(mock_urls)

    def test_scrape_data_item_urls(self):
        """Test '_scrape_data' gets item URLs from a list page."""
        mock_item_urls_response = Response()
        mock_item_urls_response.url = "url?search=1"
        mock_item_urls_response.status_code = 200

        self.pipeline.scraper.list_url_template = "url?search="
        self.pipeline.scraper.get_item_urls.return_value = []

        data = self.pipeline._scrape_data(mock_item_urls_response)
        self.assertEqual(data, [])

        self.pipeline.scraper.get_item_urls.assert_called_once_with(
            mock_item_urls_response
        )
        with self.assertRaises(AssertionError):
            self.pipeline.scraper.get_item_properties.assert_called_once_with()

    def test_scrape_data_item_properties(self):
        """Test '_scrape_data' gets item properties from an item page."""
        mock_item_response = Response()
        mock_item_response.url = "url-item-1"
        mock_item_response.status_code = 200

        self.pipeline.scraper.list_url_template = "url?search="
        self.pipeline.scraper.get_item_properties.return_value = {}

        data = self.pipeline._scrape_data(mock_item_response)
        self.assertEqual(data, {})

        self.pipeline.scraper.get_item_properties.assert_called_once_with(
            mock_item_response
        )
        with self.assertRaises(AssertionError):
            self.pipeline.scraper.get_item_urls.assert_called_once_with()

    @patch("scrapemeagain.pipeline.Pipeline._scrape_data")
    def test_actually_collect_data(self, mock_scrape_data):
        """Test '_actually_collect_data' populates 'data_queue'."""
        mock_scrape_data.return_value = {"key": "value"}

        mock_item_response = Response()
        mock_item_response.url = "url-item-1"
        mock_item_response.status_code = 200

        self.pipeline._actually_collect_data(mock_item_response)

        self.pipeline.data_queue.put.assert_called_once_with({"key": "value"})
        self.pipeline.scraping_in_progress.set.assert_called_once_with()
        self.pipeline.scraping_in_progress.clear.assert_called_once_with()

    @patch("scrapemeagain.pipeline.logging")
    @patch("scrapemeagain.pipeline.Pipeline._scrape_data")
    def test_actually_collect_data_fails(self, mock_scrape_data, mock_logging):
        """Test '_actually_collect_data' logs an exception message on fail."""
        mock_scrape_data.side_effect = ValueError

        mock_item_response = Response()
        mock_item_response.url = "url-item-1"
        mock_item_response.status_code = 200

        self.pipeline._actually_collect_data(mock_item_response)

        self.pipeline.scraping_in_progress.set.assert_called_once_with()
        self.pipeline.scraping_in_progress.clear.assert_called_once_with()
        mock_logging.error.assert_called_once_with(
            'Failed processing response for "url-item-1"'
        )
        self.assertEqual(mock_logging.exception.call_count, 1)

    @patch("scrapemeagain.pipeline.Pipeline._actually_collect_data")
    def test_collect_data(self, mock_actually_collect_data):
        """Test 'collect_data' populates and passes response bulks for
        processing.
        """
        mock_urls = ["url1", "url2", "url3"]
        mock_url_statuses = [200, 500, 200]
        mock_responses = create_responses(mock_urls, mock_url_statuses)
        self.pipeline.response_queue.get.side_effect = mock_responses + [EXIT]

        self.pipeline.collect_data()

        # 4 = len(mock_responses) + EXIT
        self.assertEqual(self.pipeline.response_queue.get.call_count, 4)

        # 3 = len(mock_responses)
        self.assertEqual(mock_actually_collect_data.call_count, 3)

    def test_store_item_urls(self):
        """Test '_store_item_urls' stores item URLs to DB."""
        mock_data = [{"url": "url1"}, {"url": "url2"}]

        self.pipeline._store_item_urls(mock_data)

        self.pipeline.databaser.insert_multiple.assert_called_once_with(
            mock_data, self.pipeline.databaser.item_urls_table
        )

    def test_store_item_urls_empty(self):
        """Test '_store_item_urls' only stores actual item URLs."""
        self.pipeline._store_item_urls([])

        self.assertEqual(self.pipeline.databaser.insert_multiple.call_count, 0)

    def test_store_item_properties(self):
        """Test '_store_item_properties' saves item properties to DB."""
        mock_data = {"url": "url1", "key": "value"}

        self.pipeline._store_item_properties(mock_data)

        self.pipeline.databaser.insert.assert_called_once_with(
            mock_data, self.pipeline.databaser.item_data_table
        )
        self.pipeline.databaser.delete_url.assert_called_once_with(
            mock_data["url"]
        )

    def test_store_item_properties_empty(self):
        """Test '_store_item_properties' does not store the item and only
        removes its URL if there are no properties set.
        """
        mock_data = {"url": "url1"}

        self.pipeline._store_item_properties(mock_data)

        self.assertEqual(self.pipeline.databaser.insert.call_count, 0)
        self.pipeline.databaser.delete_url.assert_called_once_with(
            mock_data["url"]
        )

    @patch("scrapemeagain.pipeline.Pipeline._store_item_urls")
    @patch("scrapemeagain.pipeline.Pipeline._store_item_properties")
    def test_actually_store_data(
        self, mock_store_item_properties, mock_store_item_urls
    ):
        """Test '_actually_store_data' is able to handle both item URLs and
        properties.
        """
        mock_urls = [{"url": "url1"}, {"url": "url2"}]
        self.pipeline._actually_store_data(mock_urls)
        mock_store_item_urls.assert_called_once_with(mock_urls)

        mock_properties = {"url": "url1", "key": "value"}
        self.pipeline._actually_store_data(mock_properties)
        mock_store_item_properties.assert_called_once_with(mock_properties)

    # @patch('scrapemeagain.pipeline.Pipeline._store_item_urls')
    # @patch('scrapemeagain.pipeline.Pipeline._store_item_properties')
    # def test_actually_store_data_commits(
    #     self, mock_store_item_properties, mock_store_item_urls
    # ):
    #     """Test '_actually_store_data' commits changes on exit.
    #     """
    #     self.pipeline._actually_store_data(EXIT)
    #
    #     self.pipeline.databaser.commit.assert_called_once_with()

    @patch("scrapemeagain.pipeline.logging")
    def test_actually_store_data_fails(self, mock_logging):
        """Test '_actually_store_data' logs an exception message on fail."""
        self.pipeline._actually_store_data(None)

        mock_logging.error.assert_called_once_with("Failed storing data")
        self.assertEqual(mock_logging.exception.call_count, 1)

    @patch("scrapemeagain.pipeline.Pipeline._actually_store_data")
    def test_store_data(self, mock_actually_store_data):
        """Test 'store_data' takes data from the 'data_queue' and pass it to
        '_actually_store_data' (to save it in DB), and all changes are
        committed before exiting.
        """
        mock_data = [
            {"url": "url1", "key": "value"},
            {"url": "url2", "key": "value"},
        ]
        self.pipeline.data_queue.get.side_effect = mock_data + [EXIT]

        self.pipeline.store_data()

        # 3 = len(mock_data) + EXIT
        self.assertEqual(self.pipeline.data_queue.get.call_count, 3)
        self.assertEqual(self.pipeline._actually_store_data.call_count, 2)

        # check 'urls_processed' is reset before actualy storing data
        self.assertEqual(self.pipeline.urls_processed.value, 0)

        # Should commit after EXIT.
        self.pipeline.databaser.commit.assert_called_once_with()

    @patch("scrapemeagain.pipeline.Pipeline.inform")
    def test_exit_workers(self, mock_inform):
        """Test 'exit_workers' passes an EXIT message to all queues."""
        self.pipeline.exit_workers()

        mock_inform.assert_called_once_with("Exiting workers, please wait ...")

        self.pipeline.url_queue.put.assert_called_once_with(EXIT)
        self.pipeline.response_queue.put.assert_called_once_with(EXIT)
        self.pipeline.data_queue.put.assert_called_once_with(EXIT)

    def test_queues_empty(self):
        """Test '_queues_empty' checks if all queues are empty."""
        self.pipeline.url_queue.empty.return_value = True
        self.pipeline.response_queue.empty.return_value = True
        self.pipeline.data_queue.empty.return_value = True

        queues_empty = self.pipeline._queues_empty()

        self.assertTrue(queues_empty)

        self.pipeline.url_queue.empty.assert_called_once_with()
        self.pipeline.response_queue.empty.assert_called_once_with()
        self.pipeline.data_queue.empty.assert_called_once_with()

    def test_workers_idle(self):
        """Test '_workers_idle' checks if all workers are idle."""
        self.pipeline.producing_urls_in_progress.is_set.return_value = False
        self.pipeline.requesting_in_progress.is_set.return_value = False
        self.pipeline.scraping_in_progress.is_set.return_value = False

        workers_idle = self.pipeline._workers_idle()

        self.assertTrue(workers_idle)

        self.pipeline.producing_urls_in_progress.is_set.assert_called_once_with()  # noqa
        self.pipeline.requesting_in_progress.is_set.assert_called_once_with()
        self.pipeline.scraping_in_progress.is_set.assert_called_once_with()

    @patch("scrapemeagain.pipeline.time")
    @patch("scrapemeagain.pipeline.Pipeline._inform_progress")
    @patch("scrapemeagain.pipeline.Pipeline.exit_workers")
    @patch("scrapemeagain.pipeline.Pipeline._queues_empty")
    @patch("scrapemeagain.pipeline.Pipeline._workers_idle")
    def test_switch_power(
        self,
        mock_workers_idle,
        mock_queues_empty,
        mock_exit_workers,
        mock_inform_progress,
        mock_time,
    ):
        """Test 'switch_power' repeatedly checks the program can end."""
        mock_queues_empty.side_effect = [False, False, True, True, True]
        mock_workers_idle.side_effect = [False, True, True]

        self.pipeline.switch_power()

        # Both 'url_queue' and 'response_queue' aren't empty once so we expect
        # 'switch_power' will need to wait 2 times.
        self.assertEqual(mock_time.sleep.call_count, 2)
        self.assertEqual(mock_inform_progress.call_count, 2)
        self.assertEqual(self.pipeline.url_queue.put.call_count, 0)

        mock_exit_workers.assert_called_once_with()

    @patch("scrapemeagain.pipeline.logging")
    @patch("scrapemeagain.pipeline.time")
    @patch("scrapemeagain.pipeline.Pipeline._inform_progress")
    @patch("scrapemeagain.pipeline.Pipeline.exit_workers")
    @patch("scrapemeagain.pipeline.Pipeline._queues_empty")
    @patch("scrapemeagain.pipeline.Pipeline._workers_idle")
    def test_switch_power_dumps_urls(
        self,
        mock_workers_idle,
        mock_queues_empty,
        mock_exit_workers,
        mock_inform_progress,
        mock_time,
        mock_logging,
    ):
        """Test 'switch_power' is able to dump URLs bucket."""
        mock_queues_empty.side_effect = [True, True, True]
        mock_workers_idle.side_effect = [True, True, True]
        type(self.pipeline.urls_bucket_empty).value = PropertyMock(
            side_effect=[0, 0, 1]
        )

        self.pipeline.switch_power()

        self.assertEqual(mock_time.sleep.call_count, 1)
        self.assertEqual(mock_inform_progress.call_count, 1)
        self.pipeline.url_queue.put.assert_called_once_with(DUMP_URLS_BUCKET)

        mock_logging.info.assert_called_once_with("Dumping URLs bucket")
        mock_exit_workers.assert_called_once_with()

    @patch("scrapemeagain.pipeline.Process")
    def test_employ_worker(self, mock_process):
        """Test 'employ_worker' creates and registers a dameon worker."""
        self.pipeline.employ_worker(all)

        mock_process.assert_called_once_with(target=all)
        self.assertTrue(mock_process.daemon)
        mock_process.return_value.start.assert_called_once_with()

        self.assertEqual(len(self.pipeline.workers), 1)

    @patch("scrapemeagain.pipeline.Process")
    def test_release_workers(self, mock_process):
        """Test 'release_workers' waits till a worker is finished."""
        self.pipeline.workers = [mock_process]

        self.pipeline.release_workers()

        mock_process.join.assert_called_once_with()

    @patch("scrapemeagain.pipeline.Pipeline.inform")
    @patch("scrapemeagain.pipeline.Pipeline.release_workers")
    @patch("scrapemeagain.pipeline.Pipeline.employ_worker")
    @patch("scrapemeagain.pipeline.Pipeline.get_html")
    def test_get_item_urls(
        self,
        mock_get_html,
        mock_employ_worker,
        mock_release_workers,
        mock_inform,
    ):
        """Test 'get_item_urls' starts all necessary workers and waits till
        they are finished.
        """
        self.pipeline.get_item_urls()

        mock_inform.assert_called_once_with("Collecting item URLs")

        mock_employ_worker.assert_any_call(self.pipeline.produce_list_urls)
        mock_employ_worker.assert_any_call(self.pipeline.collect_data)
        mock_employ_worker.assert_any_call(self.pipeline.store_data)
        mock_employ_worker.assert_any_call(self.pipeline.switch_power)
        mock_get_html.assert_called_once_with()
        mock_release_workers.assert_called_once_with()

    @patch("scrapemeagain.pipeline.Pipeline.inform")
    @patch("scrapemeagain.pipeline.Pipeline.release_workers")
    @patch("scrapemeagain.pipeline.Pipeline.employ_worker")
    @patch("scrapemeagain.pipeline.Pipeline.get_html")
    def test_get_item_properties(
        self,
        mock_get_html,
        mock_employ_worker,
        mock_release_workers,
        mock_inform,
    ):
        """Test 'get_item_properties' starts all necessary workers and waits
        till they are finished.
        """
        self.pipeline.get_item_properties()

        mock_inform.assert_called_once_with("Collecting item properties")

        mock_employ_worker.assert_any_call(self.pipeline.produce_item_urls)
        mock_employ_worker.assert_any_call(self.pipeline.collect_data)
        mock_employ_worker.assert_any_call(self.pipeline.store_data)
        mock_employ_worker.assert_any_call(self.pipeline.switch_power)
        mock_get_html.assert_called_once_with()
        mock_release_workers.assert_called_once_with()



"""ScrapeMeAgain Package Handler"""


from os import path
from setuptools import setup, find_packages


def read(fname, readlines=False):
    with open(path.join(path.abspath(path.dirname(__file__)), fname)) as f:
        return f.readlines() if readlines else f.read()


requirements = read("requirements.txt", True)


setup(
    version="1.0.5",
    name="scrapemeagain",
    url="https://github.com/DusanMadar/ScrapeMeAgain",
    author="Dusan Madar",
    author_email="madar.dusan@gmail.com",
    description="Yet another Python web scraping application",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    keywords="web scraper",
    packages=find_packages(),
    scripts=["scripts/scrapemeagain-compose.py"],
    include_package_data=True,
    test_suite="tests",
    license="MIT",
    platforms="linux",
    python_requires=">=3.6",
    install_requires=requirements,
    tests_require=requirements,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from flask import request

from scrapemeagain.config import Config
from scrapemeagain.dockerized.utils import app_factory, get_class_from_path


app = app_factory(__name__)


datastore_class = get_class_from_path(Config.DATASTORE_DATABASER_CLASS)
DATASTORE = datastore_class()


@app.route("/insert-data/", methods=["POST"])
def insert_data():
    DATASTORE.insert(request.json)
    return "", 201


@app.route("/commit/")
def commit():
    DATASTORE.commit()
    return "", 204


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=Config.DATASTORE_PORT)



"""Common alphanumeric functions."""


from datetime import datetime


DATE_FORMAT = "%Y.%m.%d"
TIME_FORMAT = "%H:%M:%S"
DATE_TIME_FORMAT = "{d} {t}".format(d=DATE_FORMAT, t=TIME_FORMAT)


def get_current_date():
    """Get a `YYYY.MM.DD` date stamp.

    :returns str
    """
    return datetime.strftime(datetime.now(), DATE_FORMAT)


def get_current_datetime():
    """Get a `YYYY.MM.DD hh:mm:ss` datetime stamp.

    :returns str
    """
    return datetime.strftime(datetime.now(), DATE_TIME_FORMAT)



from unittest import TestCase
from unittest.mock import Mock

from scrapemeagain.pipeline import Pipeline


class TestPipelineBase(TestCase):
    def setUp(self):
        self.pipeline = Pipeline(Mock(), Mock(), Mock())

        # Mock Scraper.
        self.pipeline.scraper.get_item_urls = Mock()
        self.pipeline.scraper.get_item_properties = Mock()

        # Mock Databaser.
        self.pipeline.databaser.insert = Mock()
        self.pipeline.databaser.insert_multiple = Mock()
        self.pipeline.databaser.delete_url = Mock()
        self.pipeline.databaser.commit = Mock()

        #
        # Mock `prepare_pipeline()`.
        # Mock pipeline's Pool and Pool.map.
        mock_pool = Mock()
        mock_pool.return_value.map = Mock()
        self.pipeline.pool = mock_pool

        # Ensure each pipeline's Queue is an unique Mock object.
        self.pipeline.url_queue = Mock()
        self.pipeline.response_queue = Mock()
        self.pipeline.data_queue = Mock()

        # Ensure each pipeline's Event is an unique Mock object.
        self.pipeline.producing_urls_in_progress = Mock()
        self.pipeline.requesting_in_progress = Mock()
        self.pipeline.scraping_in_progress = Mock()

        # Mock counter Values.
        mock_urls_to_process = Mock()
        mock_urls_to_process.value = 0
        self.pipeline.urls_to_process = mock_urls_to_process
        mock_urls_processed = Mock()
        mock_urls_processed.value = 0
        self.pipeline.urls_processed = mock_urls_processed

        # Mock URLs bucket state.
        self.pipeline.urls_bucket_empty = Mock()
